//! This module contains the code necessary to run and supervise actors.

use std::{future::Future, pin::Pin};

use log::{debug, error};
use tokio::{
    select,
    sync::{mpsc, oneshot},
    task::{AbortHandle, JoinSet},
};

use crate::ActorPanic;

pub(super) type BoxedFuture = Pin<Box<dyn Send + Future<Output = ()>>>;

pub(super) struct SpawnReq {
    task: BoxedFuture,
    sender: oneshot::Sender<AbortHandle>,
}

impl SpawnReq {
    pub(super) fn new<Fut>(task: Fut) -> (Self, oneshot::Receiver<AbortHandle>)
    where
        Fut: 'static + Send + Future<Output = ()>,
    {
        let (sender, receiver) = oneshot::channel();
        let req = Self {
            task: Box::pin(task),
            sender,
        };
        (req, receiver)
    }
}

/// The kernel is responsible for spawning and supervising actors.
pub(super) async fn kernel(mut tasks: mpsc::Receiver<SpawnReq>) {
    let mut running = JoinSet::<()>::new();

    loop {
        select! {
            option = tasks.recv() => {
                match option {
                    Some(SpawnReq { task, sender }) => {
                        let handle = running.spawn(task);
                        if sender.send(handle).is_err() {
                            error!("SpawnReq sender dropped the channel. Aborting actor.");
                        }
                    }
                    None => panic!("The Runtime was dropped!")
                };
            }
            Some(Err(join_error)) = running.join_next() => {
                match join_error.try_into_panic() {
                    Ok(panic_obj) => {
                        if let Some(actor_panic) = panic_obj.downcast_ref::<ActorPanic>() {
                            error!("{actor_panic}")
                        } else if let Some(message) = panic_obj.downcast_ref::<String>() {
                            error!("Actor panicked in an unknown state: {message}");
                        } else {
                            error!("Actor panicked for an unknown reason.");
                        }
                    }
                    Err(join_error) => {
                        debug!("Actor was aborted: {join_error}");
                    }
                };

            }
        }
    }
}
